/**
 *  GraphTFT plugin for the Video Disk Recorder 
 * 
 *  display.c - A plugin for the Video Disk Recorder
 *
 *  (c) 2004-2006 Lars Tegeler, Sascha Volkenandt, Jrg Wendel
 *
 * This code is distributed under the terms and conditions of the
 * GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
 *
 * $Id: display.c,v 1.8 2007/06/17 15:24:18 root Exp $
 * 
 **/

//***************************************************************************
// Includes
//***************************************************************************

#include <algorithm>
#include <sstream>

#include <vdr/menu.h>
#include <vdr/eitscan.h>

#include "display.h"
#include "theme.h"
#include "setup.h"

// the renderdevices

#ifdef HAVE_IMLIB 
# include "fbrenderer.h"
# include "dvbrenderer.h"
# include "dmyrenderer.h"
#endif

#ifdef HAVE_DFB
# include "dfbrenderer.h"
#endif

//***************************************************************************
// Object
//***************************************************************************

cGraphTFTDisplay::cGraphTFTDisplay()  
{
   // init renderer

   _render = 0;
   
   // init thread stuff

   _active = no;
   comThread = 0;

   // init contents

   currentSection = 0;
   lastSection = 0;
   SetTabs(0);
   charInTabs[0] = charInTabs[1] = charInTabs[2] = charInTabs[3] = charInTabs[4] = charInTabs[5] = charInTabs[6] =  0;
   _mode = NormalView;
   _sectionName = "";
   channelType = ctTv;
   _channel = 0;
   _presentTime = 0;
   _followingTime = 0;
   _volume = cDevice::CurrentVolume();
   _mute = cDevice::PrimaryDevice()->IsMute();
   _showRecording = 0;
   _replay.control = 0;
   _replay.lastMode = ModeUnknown;
   _menu.currentPosLast = 0;
   _menu.currentPos = 0;
   _menu.topPos = -1;
   _menu.lineHeight = -1;

   // calibration stuff

   calibrationPos.x = 100;
   calibrationPos.y = 100;

   // init pbp stuff

   _pbpActive = false;
}

cGraphTFTDisplay::~cGraphTFTDisplay() 
{
   Stop();

   if (comThread) delete comThread;
   if (_render) delete _render;
}

//***************************************************************************
// Init
//***************************************************************************

int cGraphTFTDisplay::Init(const char* dev, const char* cfgDir, int port)
{
   char* pos = 0;
   int devnum = -1;

#ifdef HAVE_DFB

   if ((pos = strstr(dev, "directFB")))
   {
      devnum = 0;
      _render = new DFbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                720, 576,
                                GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf, 
                                theTheme->getDir());
   }

#endif // HAVE_DFB

#ifdef HAVE_IMLIB

   if (!_render)
   {
      if ((pos = strstr(dev, "/dev/fb")))
      {
         devnum = atoi(pos+7);
         _render = new FbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                  720, 576,
                                  GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                  theTheme->getDir());
      }

      else if ((pos = strstr(dev, "vdr/")))
      {    
         devnum = atoi(pos+4);
         _render = new DvbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                   720, 576,
                                   GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                   theTheme->getDir());
      }

      else
      {
         _render = new DummyRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                     720, 576,
                                     GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                     theTheme->getDir());
      }

#ifdef WITH_TCP_COM

      comThread = new ComThread(this);

      if (comThread->init(_render, port) != 0)
      {
         d(0, "Can't establish listener, tcp communication not available");
         delete comThread;
         comThread = 0;
      }
      else
      {
         comThread->Start();
      }

#endif // WITH_TCP_COM
   }

#endif // HAVE_IMLIB

   // either the com thread or a device is needed

   if (!comThread && devnum == -1) 
      return 0;

   _render->setBorder(GraphTFTSetup.VdrWidth, GraphTFTSetup.VdrHeight);

   if (_render->init(devnum) != 0)
      return 0;

   // Show the start image

   _render->image(theTheme->getStartImage().c_str(), 0,0,0,0);
   _render->refresh();
   _render->setPlayMode(false);

   cDisplayItem::setRenderer(_render);
   cDisplayItem::setVdrStatus(this);

   Start();

   return 1;
}

//***************************************************************************
// Mode
//***************************************************************************

int cGraphTFTDisplay::setMode(DisplayMode mode, const char* menuName, int force)
{ 
   if (_mode == Calibration && !force)
      return done;
   
   if (_mode != mode || (menuName && _sectionName != menuName))
   {
      _mode = mode;
      _sectionName = menuName ? menuName : "Menu";

      d(2, "Mode is set to (0x%X), menu section to '%s'", 
        _mode, _sectionName.c_str());

      _doUpdate.Broadcast();
   }

   return done;
}

//***************************************************************************
// Setup Changed
//***************************************************************************

void cGraphTFTDisplay::setupChanged()
{
   logDevice = GraphTFTSetup.LogDevice;
   logLevel = GraphTFTSetup.Level;

   if (comThread)
      comThread->setJpegQuality(GraphTFTSetup.JpegQuality);

   if (_render)
   {
      _render->setBorder(GraphTFTSetup.VdrWidth, GraphTFTSetup.VdrHeight);
      _render->setProperties(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                             720, 576,
                             GraphTFTSetup.Iso2Utf,
                             theTheme->getDir());
   }
}

//***************************************************************************
// Set Calibrate
//***************************************************************************

void cGraphTFTDisplay::setCalibrate(int active)
{
   d(1, "Info: %s calibration mode", active ? "starting" : "stopping");

   if (active)
      setMode(cGraphTFTDisplay::Calibration);
   else
      setMode(cGraphTFTDisplay::NormalView, 0, /*force*/ true);

   _doUpdate.Broadcast();
}

//***************************************************************************
// Stop
//***************************************************************************

void cGraphTFTDisplay::Stop() 
{
   if (_active) 
   {
      _active = no;
      _doUpdate.Broadcast();
      Cancel(3);
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::Pbp() 
{
   d(1, "Pbp()");

   if (!_pbpActive)
   {
      PbpInit();
      PbpStart();
   }
   else
   {
      PbpStop();
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpRestart()
{
   d(1, "PbpRestart()");

   if (_pbpActive)
   {
      PbpStop();
      PbpStart();
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpSwitch(bool Back) 
{
   const cChannel* newChan = 0;

   d(1, "PbpSwitch()");

   if (_pbpActive)
   {
      PbpStop();

      if (!Back)
      {
         newChan = cDevice::CurrentChannel() != 0 ? Channels.GetByNumber(cDevice::CurrentChannel()) : NULL;
      }

      cDevice::PrimaryDevice()->SwitchChannel(_chan, true);

      if ( !Back )
      {
         if (newChan != NULL)
         {
            _chan = newChan;
#if APIVERSNUM >= 10500
            _dev = cDevice::GetDevice(_chan, 1, false);
#else
            _dev = cDevice::GetDevice(_chan, 1);
#endif
            PbpStart();
         }
      }
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpUpDown(bool Up) 
{
   d(1, "display.c: PbpUpDown()");

   if (_pbpActive)
   {
      // Set new Channel and device
      // but not today ;-)

      if (Up)
      {
         // set TFT Channel one up
      }
      else
      {
         // and here down
      }

      // Now make a restart

      // PbpRestart();
   }
}

//***************************************************************************
// 
//***************************************************************************

bool cGraphTFTDisplay::PbpIsRunning()
{
   d(1, "PbpIsRunning() (%s)", _pbpActive ? "yes" : "no");

   return _pbpActive;
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpInit() 
{
   d(1, "PbpInit()");

   _chan = cDevice::CurrentChannel() != 0 ? Channels.GetByNumber(cDevice::CurrentChannel()) : 0;

   if (_chan)
#if APIVERSNUM >= 10500
      _dev = cDevice::GetDevice(_chan, 1, false);
#else
      _dev = cDevice::GetDevice(_chan, 1);
#endif

      //_dev = cDevice::ActualDevice();
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpStart() 
{
   d(1, "PbpStart() dev '%s', chan '%s'", 
     _dev  ? "yes" : "no", 
     _chan ? "yes" : "no");

   if (_dev && _chan)
   {
      d(1, "Set PlayMode to Video");

      Stop();
      
      _render->setPlayMode(true);
      _dev->SwitchChannel(_chan, false);

      // active transfer

      _transfer = new cGraphTFTTransfer(_render, _chan);
      _dev->AttachReceiver(_transfer);

      _pbpActive = true;
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpStop() 
{
   d(1, "PbpStop()");

   if (_transfer) delete _transfer;

   _render->setPlayMode(false);
   _pbpActive = false;
   Start();
   _doUpdate.Broadcast();
}

//***************************************************************************
// Set Tabs
//***************************************************************************

void cGraphTFTDisplay::SetTabs(int Tab1, int Tab2, int Tab3, int Tab4, int Tab5)
{
   d(6, "SetTabs - Tab1: %d, Tab2, %d, Tab3 %d, Tab4 %d, Tab5 %d",Tab1, Tab2, Tab3, Tab4, Tab5);
   
   tabs[0] = 0;
   tabs[1] = Tab1 ? tabs[0] + Tab1 : 0;
   tabs[2] = Tab2 ? tabs[1] + Tab2 : 0;
   tabs[3] = Tab3 ? tabs[2] + Tab3 : 0;
   tabs[4] = Tab4 ? tabs[3] + Tab4 : 0;
   tabs[5] = Tab5 ? tabs[4] + Tab5 : 0;
   tabs[6] = 100;
}

//***************************************************************************
// Get Tabbed Text
//***************************************************************************

const char* cGraphTFTDisplay::GetTabbedText(const char *s, int Tab)
{
   if (!s)
      return 0;

   static char buffer[1000];
   const char *a = s;
   const char *b = strchrnul(a, '\t');

   while (*b && Tab-- > 0) 
   {
      a = b + 1;
      b = strchrnul(a, '\t');
   }

   if (!*b)
      return (Tab <= 0) ? a : 0;

   unsigned int n = b - a;

   if (n >= sizeof(buffer))
      n = sizeof(buffer) - 1;

   strncpy(buffer, a, n);
   buffer[n] = 0;

   return buffer;
}

//***************************************************************************
// Action
//***************************************************************************

void cGraphTFTDisplay::Action()
{
   uint64_t updateIn;
   int n;

   isyslog("GraphTFT plugin display thread started (pid=%d)", getpid());

   // give the plugin time to collect some data AND give vdr some time to
   // finish initialization before acquiring the lock

   usleep(500000);

   while (!theTheme->isInitialized())
      usleep(100000);

   _mutex.Lock();       // mutex gets ONLY unlocked when sleeping
   _active = yes;

   // main loop

   while (_active)
   {
      d(1, "action loop");

      if (isModeNormal(_mode))
      {
         if (GraphTFTSetup.normalMode == "Standard")
            n = Display(channelType == ctTv ? "NormalTV" : "NormalRadio");
         else
            n = Display("Normal" + GraphTFTSetup.normalMode);
      }
      else
      {
         switch (_mode) 
         {
            case ReplayNormal:      n = Display("ReplayNormal");    break;
            case ReplayMP3:         n = Display("ReplayMP3");       break;
            case ReplayDVD:         n = Display("ReplayDVD");       break;
            case Calibration:       n = Display("Calibration");     break;

            case ModeMenu:          n = Display(_sectionName);      break;

            default:                n = Display("NormalTV");
         }
      }

      updateIn = SECONDS(60);  // the default

      if (currentSection)
      {
         updateIn = currentSection->getNextUpdateTime() - cTimeMs::Now();

         // cant calc this inline below, look like a format string problem ... ? 

         int s = updateIn/1000; int us = updateIn%1000;

         d(0, "Displayed %d Items, schedule next update in %d,%03d seconds", 
           n, s, us);
      }
         
      if (updateIn < 500)
         updateIn = 500;

      _doUpdate.TimedWait(_mutex, updateIn);
   }

   isyslog("GraphTFT plugin display thread ended (pid=%d)", getpid());
}

//***************************************************************************
// Stopped Timer
//***************************************************************************

bool cGraphTFTDisplay::StoppedTimer(const char* Name)
{
   cTimer* timer = Timers.First();

   while (timer)
   {
      if (strcmp(Name, timer->File()) == 0)
         break;

      timer = Timers.Next(timer);
   }

   return timer == 0 || !timer->Recording();
}

//***************************************************************************
// Update Programme
//***************************************************************************

void cGraphTFTDisplay::UpdateProgramme()
{
   const cChannel *channel = Channels.GetByNumber(_channel);

   if (channel) 
   {
      cMutexLock lock(&_mutex);

      const cEvent *present = 0, *following = 0;
      cSchedulesLock schedulesLock;
      const cSchedules *schedules = cSchedules::Schedules(schedulesLock);

      if (schedules) 
      {
         const cSchedule *schedule = schedules->GetSchedule(channel->GetChannelID());

         if (schedule) 
         {
            if ((present = schedule->GetPresentEvent()) != 0) 
            {
               _presentTime = present->StartTime();
               _presentTitle = present->Title() ? present->Title() : "";
               _presentSubtitle = present->ShortText() ? present->ShortText() : "";
            }

            if ((following = schedule->GetFollowingEvent()) != 0) 
            {
               _followingTime = following->StartTime();
               _followingTitle = following->Title() ? following->Title() : "";
               _followingSubtitle = following->ShortText() ? following->ShortText() : "";
            }
         }
      }
   }
}

//***************************************************************************
// Osd Channel Switch
//***************************************************************************

void cGraphTFTDisplay::ChannelSwitch(const cDevice *Device, int ChannelNumber) 
{
   d(3, "ChannelSwitch on %p: %d", Device, ChannelNumber);

   if (Device->IsPrimaryDevice() && !EITScanner.UsesDevice(Device) && _channel != ChannelNumber) 
   {
      const cChannel* channel;

      cMutexLock lock(&_mutex);
      _channel = ChannelNumber;

      if ((channel = Channels.GetByNumber(_channel)) != 0)
      {
         char *temp;
         int pos;

         asprintf(&temp, "%d %s", channel->Number(), channel->Name());
         _channelText = temp;

         if ((pos = _channelText.rfind(';')) != -1)
            _channelText.erase(pos);
         if ((pos = _channelText.rfind(',')) != -1)
            _channelText.erase(pos);

         free(temp);

         if (!isModeMenu(_mode)) 
         {
            switch (channel->Vpid()) 
            {
               case 0:
               case 1:
               case 0x1fff: channelType = ctRadio;  break;
               default:     channelType = ctTv;     break;
            }

            setMode(NormalView);
         }
      }

      updateGroup(groupChannel);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Channel
//***************************************************************************

void cGraphTFTDisplay::OsdChannel(const char *Text) 
{
   d(3, "OsdChannel: %s", Text);

   if (_channelInfo != Text)
   {
      int pos;
      string temp = Text;
      cMutexLock lock(&_mutex);

      // workaround for some osd themes (Elchi)

      _channelInfo = "";
      _presentTitle = "";
      _presentSubtitle = "";
      _followingTitle = "";
      _followingSubtitle = "";

      if ((pos = temp.find(' ')) != -1) 
      {
         _channelInfo += temp.substr(0, pos) + " ";

         while (pos < (int)temp.length() && temp[pos] == ' ')
            ++pos;

         int l = temp.length();

         while (temp[l - 1] == ' ')
            --l;

         _channelInfo += temp.substr(pos, l - pos);
      } 
      else
         _channelInfo = Text;

      if ((pos = _channelInfo.rfind(';')) != -1)
         _channelInfo.erase(pos);

      if ((pos = _channelInfo.rfind(',')) != -1)
         _channelInfo.erase(pos);

      updateGroup(groupChannel);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Programme
//***************************************************************************

void cGraphTFTDisplay::OsdProgramme(time_t PresentTime, const char* PresentTitle, 
                                    const char* PresentSubtitle, time_t FollowingTime, 
                                    const char* FollowingTitle, const char* FollowingSubtitle)
{
   d(5, "OsdProgramme at %u: (Present)   %s, %s", (uint)PresentTime, PresentTitle, PresentSubtitle);
   d(5, "OsdProgramme at %u: (Following) %s, %s", (uint)FollowingTime, FollowingTitle, FollowingSubtitle);

   if (!PresentTitle)       PresentTitle = "";
   if (!PresentSubtitle)    PresentSubtitle = "";
   if (!FollowingTitle)     FollowingTitle = "";
   if (!FollowingSubtitle)  FollowingSubtitle = "";

   if (_presentTime != PresentTime || _presentTitle != PresentTitle 
       || _presentSubtitle != PresentSubtitle || _followingTime != FollowingTime 
       || _followingTitle == FollowingTitle || _followingSubtitle != FollowingSubtitle) 
   {
      cMutexLock lock(&_mutex);

      _presentTime = PresentTime;
      _presentTitle = PresentTitle;
      _presentSubtitle = PresentSubtitle;
      _followingTime = FollowingTime;
      _followingTitle = FollowingTitle;
      _followingSubtitle = FollowingSubtitle;

      updateGroup(groupChannel);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Set Volume
//***************************************************************************

void cGraphTFTDisplay::SetVolume(int, bool) 
{
   d(3, "SetVolume to %d, muted is %s", 
     cDevice::CurrentVolume(), 
     cDevice::PrimaryDevice()->IsMute() ? "true" : "false");

   if (_volume != cDevice::CurrentVolume() || _mute != cDevice::PrimaryDevice()->IsMute()) 
   {
      cMutexLock lock(&_mutex);
      _volume = cDevice::CurrentVolume();
      _mute = cDevice::PrimaryDevice()->IsMute();

      updateGroup(groupVolume);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Recording
//***************************************************************************

void cGraphTFTDisplay::Recording(const cDevice *Device, const char *Name, const char *FileName, bool On) 
{
   d(3, "Recording %s to %p", Name, Device);

   if (On) 
   {
      RecordingInfo info;

      cMutexLock lock(&_mutex);

      info.device = Device;
      info.name = Name ? Name : "";
      _recordings.push_back(info);

      updateGroup(groupRecording);
      _doUpdate.Broadcast();
   } 
   else 
   {
      cMutexLock lock(&_mutex);

      Recordings::iterator it = _recordings.begin();

      for (; it != _recordings.end(); ++it) 
      {
         if ((*it).device == Device && StoppedTimer((*it).name.c_str())) 
         {
            d(5, "deleting recording %s", (*it).name.c_str());
            _recordings.erase(it);
            break;
         }
      }

      if (_recordings.size())
         _showRecording = _showRecording % _recordings.size();

      updateGroup(groupRecording);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Replaying
//***************************************************************************

void cGraphTFTDisplay::Replaying(const cControl *Control, const char *Name, 
                                 const char *FileName, bool On) 
{
   d(3, "Replaying %s '%s' from %p", On ? "start" : "stop", Name, Control);

   if (On)
   {
      if (!Name) Name = "";

      cMutexLock lock(&_mutex);

      _replay.control = (cControl *)Control;
      _replay.lastMode = _mode;
      setMode(ReplayNormal);

      if (strlen(Name) > 6 && Name[0]=='[' && Name[3]==']' && Name[5]=='(') 
      {
         int i;

         for (i = 6; Name[i]; ++i) 
         {
            if (Name[i] == ' ' && Name[i-1] == ')')
               break;
         }

         if (Name[i]) 
         { 
            // replaying mp3

            const char* name = skipspace(Name + i);
            _replay.name = *name ? name : tr("Unknown title");
            setMode(ReplayMP3);
         }
      } 
      else if (strcmp(Name, "DVD") == 0) 
      {
         _replay.name = Name;
         setMode(ReplayDVD);
      } 
      else if (strlen(Name) > 7) 
      {
         int i, n;

         for (i = 0, n = 0; Name[i]; ++i) 
         {
            if (Name[i] == ' ' && Name[i-1] == ',' && ++n == 4)
               break;
         }

         if (Name[i]) 
         { 
            // replaying DVD

            const char *name = skipspace(Name + i);
            _replay.name = *name ? name : tr("Unknown title");
            replace(_replay.name.begin(), _replay.name.end(), '_', ' ');
            setMode(ReplayDVD);
         }
      }

      if (_mode == ReplayNormal) 
      {
         int pos;
         _replay.name = Name;
         _replay.folder = Name;

         if ((pos = _replay.name.rfind('~')) != -1) 
         {
            _replay.name.erase(0, pos + 1);
            _replay.folder.erase(pos);
         }
      }

      updateGroup(groupReplay);
      _doUpdate.Broadcast();
   } 

   else 
   {
      cMutexLock lock(&_mutex);
      _replay.control = 0;
      _replay.name = "";
      _replay.folder = "";
 
      setMode(NormalView);

      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Status Message
//***************************************************************************

void cGraphTFTDisplay::OsdStatusMessage(const char* Message) 
{
   d(3, "OsdStatusMessage %s", Message);

   cMutexLock lock(&_mutex);

   _message = Message ? Message : "";
   
   updateGroup(groupMessage);
   _doUpdate.Broadcast();
}

//***************************************************************************
// Osd Title
//***************************************************************************

void cGraphTFTDisplay::OsdTitle(const char* Title) 
{
   d(5, "OsdTitle %s", Title);

   if (Title)
   {
      int pos;
      cMutexLock lock(&_mutex);

      _message = "";
      _menu.title = Title;

      if ((pos = _menu.title.find('\t')) != -1)
         _menu.title.erase(pos);
   }
}

//***************************************************************************
// Osd Item
//***************************************************************************

void cGraphTFTDisplay::OsdItem(const char* Text, int Index) 
{
   d(3, "OsdItem %s at %d", Text, Index);

   if (Text)
   {
      cMutexLock lock(&_mutex);
      MenuItem item;
      item.text = Text;
      item.type = itNormal;
      item.tabCount = 0;

      _message = "";

      for (int i = 0; i < MaxTabs; ++i) 
      {
         const char* tp = GetTabbedText(Text, i);

         if (tp)
         {
            char* tab;
            int len;

            tab = strdup(tp);
            strreplace(tab, '\177', '|');
            strreplace(tab, '\203', ' ');
            strreplace(tab, '\204', ' ');
            strreplace(tab, ']', ' ');
            strreplace(tab, '[', ' ');

            len = strlen(Str::allTrim(tab));

            if (charInTabs[i] < len)
               charInTabs[i] = len;

            // detect partingLine
            // some special handling for lines of epgsearch plugin like
            // ---------------------------------- \t Die 24.10.2006 ------------------------

            if (i == 1 && strstr(item.tabs[0].c_str(), "-----"))
            {
               item.type = itPartingLine;
               char* p;

               if ((p = strstr(tab, " ---")))
                  *p = 0;

               if (!strstr(tab, "-----"))
                  item.tabs[0] = Str::allTrim(tab);
               else
                  item.tabs[0] = "";

               d(2, "Detected parting line '%s'", item.tabs[0].c_str());
            }
            else
            {
               item.tabs[i] = tab;
            }

            item.tabCount++;
            free(tab);
         }
         else
            break;
      }
      
      _menu.items.push_back(item);
   }
}

//***************************************************************************
// Osd Current Item
//***************************************************************************

void cGraphTFTDisplay::OsdCurrentItem(const char* Text)
{
   d(4, "OsdCurrentItem %s", Text);

   if (!Text)
      return ;

   string text = Text;

   if (text != _menu.current)
   {
      int i;

      cMutexLock lock(&_mutex);

      _message = "";
      _menu.current = text;

      // lookup the item in the list and set the new currenzt position

      for (i = 0; i < (int)_menu.items.size(); i++) 
      {
         if (_menu.items[i].text == _menu.current)
         {
            _menu.currentPos = i;
            updateGroup(groupMenu);
            _doUpdate.Broadcast();
            return ;
         }
      }

      // We don't have found the item, resume old one with changed values ...

      if (_menu.currentPos >= (int)_menu.items.size())
         return ;

      d(2, "OsdCurrentItem: Text of current item changed to %s", Text);

#if __GNUC__ < 3
      MenuItem item = _menu.items[_menu.currentPos];
#else
      MenuItem item = _menu.items.at(_menu.currentPos);
#endif

      vector<MenuItem>::iterator tempIterator;
      tempIterator = _menu.items.begin();

      for (int i = 0; i < _menu.currentPos; i++)
         tempIterator++;

      tempIterator = _menu.items.erase(tempIterator);
      item.text = Text;

      for (int i = 0; i < MaxTabs; ++i) 
      {
         const char* tab = GetTabbedText(Text, i);

         if (tab)
            item.tabs[i] = tab;

         if (GetTabbedText(Text, i+1) == 0)
            break;
      }

      _menu.items.insert(tempIterator, item);
      updateGroup(groupMenu);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Clear
//***************************************************************************

void cGraphTFTDisplay::OsdClear()
{
   d(3, "OsdClear");

   cMutexLock lock(&_mutex);

   _menu.title = "";
   _menu.items.clear();
   _menu.text = "";

   charInTabs[0] = charInTabs[1] = charInTabs[2] = 
      charInTabs[3] = charInTabs[4] = charInTabs[5] = 0;
   SetTabs(0,0,0,0,0);

   if (_channelInfo != "") 
   {
      _channelInfo = "";
      updateGroup(groupChannel);
      _doUpdate.Broadcast();
   }

   if (_message != "")
   {
      _message = "";
      updateGroup(groupMessage);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Help Keys
//***************************************************************************

void cGraphTFTDisplay::OsdHelpKeys(const char* Red, const char* Green, 
                                   const char* Yellow, const char* Blue) 
{
   d(3, "OsdHelpKeys: '%s', '%s', '%s', '%s'", Red, Green, Yellow, Blue);

   cMutexLock lock(&_mutex);

   string red = Red ? Red : "";
   string green = Green ? Green : "";
   string yellow = Yellow ? Yellow : "";
   string blue = Blue ? Blue : "";

   if (_menu.buttons[0] != red ||
       _menu.buttons[1] != green ||
       _menu.buttons[2] != yellow ||
       _menu.buttons[3] != blue)
   {
      _menu.buttons[0] = red;
      _menu.buttons[1] = green;
      _menu.buttons[2] = yellow;
      _menu.buttons[3] = blue;
      
      updateGroup(groupButton);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Text Item
//***************************************************************************

void cGraphTFTDisplay::OsdTextItem(const char* Text, bool Scroll) 
{
   d(5, "OsdTextItem: %s scroll %s", Text, Scroll ? "true" : "false");

   cMutexLock lock(&_mutex);

   if (Text)
   { 
      _menu.text = Text;
      replace(_menu.text.begin(), _menu.text.end(), '\n', ' ');
   }

   _doUpdate.Broadcast();
}

//***************************************************************************
// Osd Menu Display
//***************************************************************************

void cGraphTFTDisplay::OsdMenuDisplay(const char* kind)
{
   d(3, "OsdMenuDisplay - kind %d", kind);

   cMutexLock lock(&_mutex);

   if (!isModeMenu(_mode)) 
      _menu.lastMode = _mode;
   
   setMode(ModeMenu, kind);
}

//***************************************************************************
// Osd Menu Destroy
//***************************************************************************

void cGraphTFTDisplay::OsdMenuDestroy()
{
   d(3, "OsdMenuDestroy");

   cMutexLock lock(&_mutex);

   // all menues closed !

   if (isModeMenu(_mode))        // 'if' is needed for replay via extrecmenu
      setMode(_menu.lastMode);      
}

//***************************************************************************
// Mouse Event
//***************************************************************************

void cGraphTFTDisplay::mouseEvent(int x, int y, int button, int flag)
{
   if (!currentSection)
      return ;

   cMutexLock lock(&_mutex);

   if (isModeCallibration(_mode))
   {
      calibrationPos.x = x;
      calibrationPos.y = y;

      _doUpdate.Broadcast();
   }

   else if (button == cGraphTftComService::mbWheelUp)
   {
      cRemote::Put(cKey::FromString("Up"));
   }

   else if (button == cGraphTftComService::mbWheelDown)
   {
      cRemote::Put(cKey::FromString("Down"));
   }

   else if (button == cGraphTftComService::mbRight)
   {
      cRemote::Put(cKey::FromString("Back"));
   }

   else if (button == cGraphTftComService::mbLeft)
   {
      for (cDisplayItem* p = currentSection->getItems()->First(); p; 
           p = currentSection->getItems()->Next(p))
      {
         // check click coordinates 
         
         if (x >= p->X() && x <= p->X()+p->Width() &&
             y >= p->Y() && y <= p->Y()+p->Height())
         {
            d(2, "item %d on (%d/%d) clicked, sending key '%s', %sfound in keytab", p->Item(),
              x, y, p->OnClick().c_str() ? p->OnClick().c_str() : "<none>",
              p->OnClick().c_str() && cKey::FromString(p->OnClick().c_str()) ? "" : "not ");

            // menue navigation area ?
            
            if (p->Item() == itemMenuNavigationArea &&
                _menu.lineHeight > 0)
            {
               int yOff = y - p->Y();
               int clickRow = yOff / _menu.lineHeight + _menu.topPos;
               
               if (_menu.currentPos < clickRow)        // down
                  for (int i = _menu.currentPos; i < clickRow; i++)
                     cRemote::Put(cKey::FromString("Down"));
               else if (_menu.currentPos > clickRow)   // up
                  for (int i = _menu.currentPos; i > clickRow; i--)
                     cRemote::Put(cKey::FromString("Up"));
            }
            
            // check if item with key definition and double clicked
            
            if (p->OnClick() != "" && flag == 0)
            {
               d(2, "item %d clicked, sending key '%s'", p->Item(), 
                 p->OnClick().c_str());

               cRemote::Put(cKey::FromString(p->OnClick().c_str()));
            }

            if (p->OnDblClick() != "" && (flag & cGraphTftComService::efDoubleClick))
            {
               d(2, "item %d double-clicked, sending key '%s'", p->Item(), 
                 p->OnDblClick().c_str());
               
               cRemote::Put(cKey::FromString(p->OnDblClick().c_str()));
            }
         }
      }
   }
}

//***************************************************************************
// Display
//***************************************************************************

int cGraphTFTDisplay::Display(string sectionName)
{
   int count = 0;

   if (isModeNormal(_mode))
      UpdateProgramme();

   else if (isModeMenu(_mode) && !theTheme->getSection(sectionName))
   {
      d(2, "Info: Section faked to '%s' due to section '%s' not defined!", 
        "Menu", sectionName.c_str());

      sectionName = "Menu";
   }

   if (!(currentSection = theTheme->getSection(sectionName)))
      return 0;

   // reset force flag

   cDisplayItem::setForce(no);

   if (currentSection != lastSection)
   {
      d(0, "Section changed from '%s' to '%s'", 
        lastSection ? lastSection->getName().c_str() : "<none>", 
        currentSection->getName().c_str());

      lastSection = currentSection;
      _menu.topPos = na;
      _menu.lineHeight = na;
      
      DisplayClear();
      cDisplayItem::setForce(yes);
   }

   // for now ...

   if (isModeMenu(_mode))
      cDisplayItem::setForce(yes);
      
   // draw items

   for (cDisplayItem* p = currentSection->getItems()->First(); p; 
        p = currentSection->getItems()->Next(p))
   {
      count += p->refresh();
   }
   
   // update display

   if (count)
      DisplayShow();

   return count;
}

void cGraphTFTDisplay::DisplayClear()
{
   _render->clear();
}

void cGraphTFTDisplay::DisplayShow()
{
   _render->refresh();

   if (comThread)
      comThread->refresh();

   // dump to file ...

   if (GraphTFTSetup.DumpImage) 
   {
      static cTimeMs timeMs;
      static int _Dumped = timeMs.Now();

      int lastDumpDiff = _Dumped ? timeMs.Now() - _Dumped : 0;

      if (lastDumpDiff > (GraphTFTSetup.DumpRefresh * 1000))
      {
         _Dumped = timeMs.Now();
         _render->dumpImage2File("/graphTFT.png", 
                                 GraphTFTSetup.DumpImageX, 
                                 GraphTFTSetup.DumpImageY);
      }
   }
}

int cGraphTFTDisplay::updateGroup(int group)
{
   if (!currentSection)
      return ignore;

   return currentSection->updateGroup(group);
}
