/**
 *  GraphTFT plugin for the Video Disk Recorder 
 * 
 *  display.h - A plugin for the Video Disk Recorder
 *
 *  (c) 2004-2006 Lars Tegeler, Sascha Volkenandt, Jrg Wendel
 *
 * This code is distributed under the terms and conditions of the
 * GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
 *
 * $Id: display.h,v 1.8 2007/06/17 15:24:18 root Exp $
 *
 **/

#ifndef __GTFT_DISPLAY_H
#define __GTFT_DISPLAY_H

//***************************************************************************
// Includes
//***************************************************************************

#include <map>
#include <string>
#include <vector>

#include <vdr/thread.h>
#include <vdr/status.h>
#include <vdr/tools.h>

#include <common.h>
#include <comthread.h>
#include <renderer.h>
#include <theme.h>
#include <transfer.h>

using std::map;
using std::string;
using std::vector;

//***************************************************************************
// Class cGraphTFTService
//***************************************************************************

class cGraphTFTService : public cThemeService
{
   public:

      enum ChannelType
      {
         ctRadio = 0,
         ctTv
      };
};

//***************************************************************************
// Class GraphTFT Display
//***************************************************************************

class cGraphTFTDisplay : public cStatus, cThread, cGraphTFTService
{
   public:

      enum { MaxTabs = 6 };
      
      enum DisplayMode 
      {
         ModeUnknown      = 0x00,

         ModeMask         = 0xF0,
         DisplayMask      = 0x0F,

         ModeNormal       = 0x10,
         ModeReplay       = 0x20,
         ModeMenu         = 0x30,
         ModeCallibration = 0x40,

         NormalView       = 0x11,

         ReplayNormal     = 0x21,
         ReplayMP3        = 0x22,
         ReplayDVD        = 0x23,

         MenuDefault      = 0x31,
         MenuMain         = 0x32,
         MenuSchedule     = 0x33,
         MenuChannels     = 0x34,
         MenuTimers       = 0x35,
         MenuRecordings   = 0x36,
         MenuSetup        = 0x37,
         MenuCommands     = 0x38,
         MenuWhatsOnElse  = 0x39,
         MenuWhatsOnNow   = 0x3a,
         MenuWhatsOnNext  = 0x3b,
         MenuSetupPage    = 0x3c,

         Calibration      = 0x41
      };

      int setMode(DisplayMode mode, const char* menuName = 0, int force = false);

      int isModeNormal(DisplayMode mode)       { return (mode & ModeMask) == ModeNormal; }
      int isModeReplay(DisplayMode mode)       { return (mode & ModeMask) == ModeReplay; }
      int isModeMenu(DisplayMode mode)         { return (mode & ModeMask) == ModeMenu; }
      int isModeCallibration(DisplayMode mode) { return (mode & ModeMask) == ModeCallibration; }


      // object

      cGraphTFTDisplay();
      ~cGraphTFTDisplay();

      // interface

      int Init(const char* dev, const char* cfgDir, int port);
      void setupChanged();
      void setCalibrate(int active);
   
      // thread stuff

      bool Start()        { return cThread::Start(); }
      void Stop();
      bool Active() const { return _active; }
      cMutex* getMutex()  { return &_mutex; }

      void themesReloaded() { lastSection = 0; }

      // due to comThread destructor is called by vdr, 
      // trick to inherited by cRemote

      void clearComThread() { comThread = 0; } 

      // event
      
      void mouseEvent(int x, int y, int button, int flag = ComThread::efNone);

      // PbP stuff

      void Pbp();
      void PbpRestart();
      void PbpSwitch(bool Back);
      void PbpUpDown(bool Up);
      bool PbpIsRunning();
      void PbpStart();
      void PbpStop();
      void PbpInit();

      // display stuff

      int Display(string sectionName);
      void DisplayShow();
      void DisplayClear();
      int updateGroup(int group);

      // Tabs

      int Tab(int n) { return (n >= 0 && n <= MaxTabs) ? tabs[n] : 0; }
      const char* GetTabbedText(const char* s, int Tab);
      virtual void SetTabs(int Tab1, int Tab2 = 0, int Tab3 = 0, int Tab4 = 0, int Tab5 = 0);

      // some structures and types

      typedef vector <string> Strings;

      struct RecordingInfo 
      {
         string name;
         const cDevice *device;
      };

      typedef vector <RecordingInfo> Recordings;
   
      struct ReplayInfo 
      {
         DisplayMode lastMode;
         string name;
         string folder;
         cControl* control;
      };

      struct CalibrationPos
      {
         int x;
         int y;
      };

      struct MenuItem 
      {
         string text;
         string tabs[MaxTabs];
         int tabCount;
         Ts::MenuItemType type;
      };
   
      struct MenuInfo 
      {
         DisplayMode lastMode;
         int currentPos;
         int currentPosLast;
         int topPos;
         int lineHeight;
         string title;
         string current;
         string buttons[4];
         string text;
         vector<MenuItem> items;
      };

      // display contents

      cThemeSection* currentSection;
      cThemeSection* lastSection;
      ChannelType channelType;
      DisplayMode _mode;
      string _sectionName;

      int _channel;

      bool _mute;
      int _volume;

      int _showRecording;

      time_t _presentTime;
      string _presentTitle;
      string _presentSubtitle;
      time_t _followingTime;
      string _followingTitle;
      string _followingSubtitle;
      string _channelText;
      string _channelInfo;

      string _message;
      Recordings _recordings;
      ReplayInfo _replay;
      MenuInfo _menu;
      CalibrationPos calibrationPos;

      int tabs[MaxTabs+1];
      int charInTabs[MaxTabs+1];

   protected:

      // thread

      virtual void Action();

      // helper functions

      bool StoppedTimer(const char *Name);
      void UpdateProgramme();

      // status interface

      virtual void ChannelSwitch(const cDevice *Device, int ChannelNumber);
      virtual void OsdChannel(const char *Text);
      virtual void OsdProgramme(time_t PresentTime, const char *PresentTitle, const char *PresentSubtitle, time_t FollowingTime, const char *FollowingTitle, const char *FollowingSubtitle);
      virtual void SetVolume(int Volume, bool Absolute);
      virtual void Replaying(const cControl *Control, const char *Name, const char *FileName, bool On);
      virtual void Recording(const cDevice *Device, const char *Name, const char *FileName, bool On);
      virtual void OsdStatusMessage(const char *Message);
      virtual void OsdTitle(const char* Title);
      virtual void OsdItem(const char* Text, int Index);
      virtual void OsdCurrentItem(const char *Text);
      virtual void OsdClear();
      virtual void OsdHelpKeys(const char* Red, const char* Green, const char* Yellow, const char* Blue);
      virtual void OsdTextItem(const char* Text, bool Scroll);
      virtual void OsdMenuDisplay(const char* kind);
      virtual void OsdMenuDestroy();

   private:

      // renderer

      Renderer* _render;
      ComThread* comThread;

      // thread control

      bool _active;
      cMutex _mutex;
      cCondVar _doUpdate;

      // pbp stuff

      bool _pbpActive;
      cDevice* _dev;
      const cChannel* _chan;
      cGraphTFTTransfer* _transfer;

};

//***************************************************************************
#endif // __GTFT_DISPLAY_H
